from __future__ import print_function
import unittest
import numpy as np
import pyqg

class VerticalModesTester(unittest.TestCase):

    def setUp(self):

        self.m = pyqg.LayeredModel(
                    nz = 2,
                    U  = [.1,.05],
                    V  = [.1,.0],
                    H  = [1.,1.],
                    delta = 1.,
                    rd = 5.,
                    f  = 1.,
                    beta = 0.,)

        self.atol=1.e-16

        self.m.set_q(np.random.randn(self.m.nz,self.m.ny,self.m.nx))

        self.m.vertical_modes()

    def test_radii(self):
        """ Check deformation radii are computed correctly """


        radii = np.array([np.sqrt(self.m.g*self.m.H)/np.abs(self.m.f), self.m.rd])
        np.testing.assert_allclose(self.m.radii,radii,atol=self.atol,
                err_msg=' Wrong deformation radii')

    def test_modes(self):
        """ Check if vertical modes are computes and normalized correctly """

        p = (self.m.Hi[:,np.newaxis]*self.m.pmodes)/self.m.H

        print(p)

        # barotropic mode must be depth invariant
        np.testing.assert_allclose(p[0,0],p[1,0],atol=self.atol,
                err_msg=' Barotropic mode is not depth invariant')

        # baroclinic modes should have zero integral
        np.testing.assert_allclose(p[0,1]+p[1,1],0.,atol=self.atol,
                err_msg=' Baroclinic mode does not integrate to zero')

        # check normalization
        np.testing.assert_allclose(p[0,0]+p[1,0],1.,atol=self.atol,
                err_msg=' Wrong normalization')


    def test_projection(self):
        """ Check modal projection """

        qn = self.m.modal_projection(self.m.qh,forward=True)
        qh = self.m.modal_projection(qn, forward=False)

        np.testing.assert_allclose(self.m.qh,qh,atol=self.atol,
                err_msg=' Modal projection does not satisfy inversion')


if __name__ == "__main__":
    unittest.main()



from __future__ import print_function
from future.utils import iteritems
import unittest
import numpy as np
import pyqg
from pyqg import diagnostic_tools as diag

class ReferenceSolutionsTester(unittest.TestCase):

    def test_two_layer(self):
        """ Tests against some statistics of a reference two-layer solution """

        year = 360*86400.
        m = pyqg.QGModel(
            nx=32,
            L=1e6,
            beta=1.5e-11,
            rek=5.787e-7,
            rd=30000.0,
            delta=0.25,
            U1=0.05,
            U2=0.0,
            filterfac=18.4,
            dt=12800.,
            tmax=3*year,
            tavestart=1*year,
            taveint=12800.,
            useAB2=True,
            diagnostics_list='all'
            )

        m.set_q1q2(
                (1e-6*np.cos(2*5*np.pi * m.x / m.L) +
                 1e-7*np.cos(2*5*np.pi * m.y / m.W)),
                np.zeros_like(m.x) )

        m.run()

        q1 = m.q[0]
        q1norm = (q1**2).sum()

        assert m.t == 93312000.0


        # machine + numpy.version fluctuations
        #   appears to be less than 0.1%
        rtol, atol = 0.01, 0.

        # These numbers come from Malte's original 2-layer simulation
        np.testing.assert_allclose(q1norm, 9.561430503712755e-08, rtol=rtol, atol=atol,
                    err_msg= ' Inconsistent with reference solution')

        diagnostic_results = {
                'APEgen': 2.5225558013107688e-07  / (m.nx**2),
                'EKEdiss': 1.4806764171539711e-07 / (m.nx**2),
               }

        sum_diagnostic_results = {
             'EKE': 0.00818377631732826 + 0.00015616609033468579,
                }

        # need to average these diagnostics
        avg_diagnostic_results = {
             'entspec': 1.5015983257921716e-06,
              'APEflux': 0.00017889483037254459,
              'KEflux':  0.00037067750708912918,
              'APEgenspec': 0.00025837684260178754,
              'KEspec': 2 * (8581.3114357188006 + 163.75201433878425) / (m.M**2)
        }


        # first print all output
        for name, des in iteritems(diagnostic_results):
            res = m.get_diagnostic(name)
            print('%10s: %1.15e \n%10s  %1.15e (desired)' % (name, res, '', des))
        for name, des in iteritems(sum_diagnostic_results):
            res = m.get_diagnostic(name).sum()
            print('%10s: %1.15e \n%10s  %1.15e (desired)' % (name, res, '', des))
        for name, des in iteritems(avg_diagnostic_results):
            res = diag.spec_sum(np.abs(m.get_diagnostic(name))).sum()
            print('%10s: %1.15e \n%10s  %1.15e (desired)' % (name, res, '', des))

        # now do assertions
        for name, des in iteritems(diagnostic_results):
            res = m.get_diagnostic(name)
            np.testing.assert_allclose(res, des, rtol=rtol, atol=atol)
        for name, des in iteritems(sum_diagnostic_results):
            res = m.get_diagnostic(name).sum()
            np.testing.assert_allclose(res, des, rtol=rtol, atol=atol)
        for name, des in iteritems(avg_diagnostic_results):
            res = diag.spec_sum(np.abs(m.get_diagnostic(name))).sum()
            np.testing.assert_allclose(res, des, rtol=rtol, atol=atol)


    def test_bt(self):
        """ Tests against some statistics of a reference barotropic solution """

        m = pyqg.BTModel(L=2.*np.pi,nx=64, tmax = 5,
                beta = 0., H = 1., rek = 0., rd = None, dt = 0.0025,
                twrite=1000)

        # IC
        p = np.exp(-(2.*(m.x-1.75*np.pi/2))**2.-(2.*(m.y-np.pi))**2) +\
                np.exp(-(2.*(m.x-2.25*np.pi/2))**2.-(2.*(m.y-np.pi))**2)

        ph = m.fft(p[np.newaxis,...])
        KEaux = m.spec_var(m.filtr*m.wv*ph )/2.
        pih = ( ph/np.sqrt(KEaux) )
        qih = -m.wv2*pih
        qi = m.ifft(qih)
        m.set_q(qi)

        rtol = 1.e-5
        atol = 1.e-14

        np.testing.assert_allclose(m.q, qi, rtol, atol)

        m.run()

        qnorm = (m.q**2).sum()
        mp = m.ifft(m.ph)
        pnorm = (mp**2).sum()
        ke = m._calc_ke()

        print('time:       %g' % m.t)
        assert m.t == 5.000000000000082

        np.testing.assert_allclose(qnorm, 89101.741238768518, rtol, atol,
                err_msg= ' Inconsistent with reference solution')
        np.testing.assert_allclose(pnorm, 1493.217664248918, rtol, atol,
                err_msg= ' Inconsistent with reference solution')
        np.testing.assert_allclose(ke, 0.9950360837282386, rtol, atol,
                err_msg= ' Inconsistent with reference solution')


    def test_sqg(self):
        """ Tests against some statistics of a reference sqg solution """

        m = pyqg.SQGModel(L=2.*np.pi,nx=64, tmax = 5.,
                beta = 0., H = 1., rek = 0., dt = 0.0025,
                twrite=1000)

        p = np.exp(-(2.*(m.x-1.75*np.pi/2))**2.-(2.*(m.y-np.pi))**2) +\
                np.exp(-(2.*(m.x-2.25*np.pi/2))**2.-(2.*(m.y-np.pi))**2)

        ph = m.fft(p[np.newaxis,:,:])
        KEaux = m.spec_var( m.filtr*m.wv*ph )/2.
        ph = ( ph/np.sqrt(KEaux) )
        qih = m.wv*ph
        qi = m.ifft(qih)
        m.set_q(qi)

        rtol = 1.e-5
        atol = 1.e-14

        np.testing.assert_allclose(m.q, qi, atol)

        m.run()

        qnorm = (m.q**2).sum()
        mp = m.ifft(m.ph)
        pnorm = (mp**2).sum()
        ke = m._calc_ke()

        print('time:       %g' % m.t)
        assert m.t == 5.000000000000082

        np.testing.assert_allclose(qnorm, 7847.5169609881032, rtol, atol,
                err_msg= ' Inconsistent with reference solution')
        np.testing.assert_allclose(pnorm, 1346.8874163575524, rtol, atol,
                err_msg= ' Inconsistent with reference solution')
        np.testing.assert_allclose(ke, 0.9579493831013508, rtol, atol,
                err_msg= ' Inconsistent with reference solution')


if __name__ == "__main__":
    unittest.main()



import numpy as np
import xarray as xr
from pyqg.errors import DiagnosticNotFilledError

# Define dict for variable dimensions
spatial_dims = ('time','lev','y','x')
spectral_dims = ('time','lev','l','k')
dim_database = {
    'q': spatial_dims,
    'u': spatial_dims,
    'v': spatial_dims,
    'ufull': spatial_dims,
    'vfull': spatial_dims, 
    'qh': spectral_dims,
    'uh': spectral_dims,
    'vh': spectral_dims,
    'ph': spectral_dims, 
    'Ubg': ('lev'),
    'Qy': ('lev'),
}

# dict for variable dimensions
var_attr_database = {
    'q': {'long_name': 'potential vorticity in real space', 'units': 'meters squared Kelvin per second per kilogram',},
    'u': {'long_name': 'zonal velocity anomaly', 'units': 'meters per second',},
    'v': {'long_name': 'meridional velocity anomaly', 'units': 'meters per second',},
    'ufull': {'long_name': 'zonal full velocities in real space', 'units': 'meters per second',},
    'vfull': {'long_name': 'meridional full velocities in real space', 'units': 'meters per second',},
    'qh': {'long_name': 'potential vorticity in spectral space', 'units': 'meters squared Kelvin per second per kilogram',},
    'uh': {'long_name': 'zonal velocity anomaly in spectral space', 'units': 'meters per second',},
    'vh': {'long_name': 'meridional velocity anomaly in spectral space', 'units': 'meters per second',},
    'ph': {'long_name': 'streamfunction in spectral space', 'units': 'meters squared per second',},
    'Ubg': {'long_name': 'background zonal velocity', 'units': 'meters per second',},
    'Qy': {'long_name': 'background potential vorticity gradient', 'units': 'meters squared Kelvin per second per kilogram',} ,
 
}

# dict for coordinate dimensions
coord_database = {
    'time': ('time'),
    'lev': ('lev'),
    'lev_mid': ('lev_mid'),
    'x': ('x'),
    'y': ('y'),
    'l': ('l'),
    'k': ('k'),
}

# dict for coordinate attributes 
coord_attr_database = {
    'time': {'long_name': 'model time', 'units': 'seconds',},
    'lev': {'long_name': 'vertical levels',},
    'lev_mid': {'long_name': 'vertical level interface',},
    'x': {'long_name': 'real space grid points in the x direction', 'units': 'grid point',},
    'y': {'long_name': 'real space grid points in the y direction', 'units': 'grid point',},
    'l': {'long_name': 'spectal space grid points in the l direction', 'units': 'meridional wavenumber',},
    'k': {'long_name': 'spectal space grid points in the k direction', 'units': 'zonal wavenumber',},
}

# list for dataset attributes
attribute_database = [
    'beta',
    'delta',
    'del2',
    'dt',
    'filterfac',
    'L',
    'M',
    'nk',
    'nl',
    'ntd',
    'nx',
    'ny',
    'nz',
    'pmodes',
    'radii',
    'rd',
    'rho',
    'rek',
    'taveint',
    'tavestart',
    'tc',
    'tmax',
    'tsnapint',
    'tsnapstart',
    'twrite',
    'W',
]

# Transform certain key coordinates
transformations = {
    'time': lambda x: np.array([x.t]),
    'lev': lambda x: np.arange(1,x.nz+1),
    'lev_mid': lambda x: np.arange(1.5,x.nz+.5),
    'x': lambda x: x.x[0,:],
    'y': lambda x: x.y[:,0],
    'l': lambda x: x.ll,
    'k': lambda x: x.kk,
}

def model_to_dataset(m):
    '''Convert outputs from model to an xarray dataset'''

    # Create a dictionary of variables
    variables = {}
    for vname in dim_database:
        if hasattr(m,vname):
            data = getattr(m,vname, None).copy()
            if 'time' in dim_database[vname]:
                variables[vname] = (dim_database[vname], data[np.newaxis,...], var_attr_database[vname])
            else:
                variables[vname] = (dim_database[vname], data, var_attr_database[vname])

    # Create a dictionary of coordinates
    coordinates = {}
    for cname in coord_database:
        data = transformations[cname](m).copy()
        coordinates[cname] = (coord_database[cname], data, coord_attr_database[cname])

    # Create a dictionary of global attributes
    global_attrs = {}
    for aname in attribute_database:
        if hasattr(m, aname):
            data = getattr(m, aname)
            global_attrs[f"pyqg:{aname}"] = (data)
        
    diagnostics = {}
    for diag_name in m.diagnostics:
        try:
            dims = m.diagnostics[diag_name]['dims']
            data = m.get_diagnostic(diag_name)
            if isinstance(data, np.float64):
                data = np.array([m.get_diagnostic(diag_name)])
            attrs = {'long_name': m.diagnostics[diag_name]['description'], 'units': m.diagnostics[diag_name]['units'],}
            diagnostics[diag_name] = (dims, data, attrs)
        except DiagnosticNotFilledError:
            pass
    
    variables.update(diagnostics)
    
    ds = xr.Dataset(variables, coords=coordinates, attrs=global_attrs)
    ds.attrs['title'] = 'pyqg: Python Quasigeostrophic Model'
    ds.attrs['reference'] = 'https://pyqg.readthedocs.io/en/latest/index.html'
    
    return ds



from __future__ import print_function
import numpy as np
from numpy import pi
import time
import pytest

# hack to skip tests if we don't have pyfft
try:
    import pyfftw
    skip=False
except ImportError:
    skip=True

@pytest.mark.skipif(skip, reason="Can't test fftw without pyfftw")
def test_fftw_rfft2(Nx = 64, Ny = None, n = 7200):
    """ A verification of rfft2/irfft2 pyfftw accuracy lost... """

    # settings
    if Ny is None: Ny = Nx

    # set random stream function
    A = np.random.randn(Nx,Ny)
    Ai = A.copy()

    tstart = time.time()
    for i in range(n):
        Ah = pyfftw.interfaces.numpy_fft.rfft2(Ai, threads=1)
        Ai = pyfftw.interfaces.numpy_fft.irfft2(Ah, threads=1)
    tend = time.time()

    # error after nmax fft cycles
    abs_err = np.abs((A-Ai)).max()
    rel_err = np.abs((A-Ai)/A).max()

    print("RFFT2 test with Nx = %i, Ny = %i" %(Nx,Ny))
    print("Performed %i Forward/Inverse FFTs in %f5 seconds, CPU time" %(n,tend-tstart))
    print("Absolute error after %i Forward/Inverse FFTs = %e" %(n,abs_err))
    print("Relative error after %i Forward/Inverse FFTs = %e " %(n,rel_err))
    print(" ")

@pytest.mark.skipif(skip, reason="Can't test fftw without pyfftw")
def test_fftw_rfft(Nx = 64, n = 100000):
    """ A verification of rfft/irfft accuracy lost """

    # set random stream function
    A = np.random.randn(Nx)
    Ai = A.copy()

    tstart = time.time()
    for i in range(n):
        Ah = pyfftw.interfaces.numpy_fft.rfft(Ai, threads=1)
        Ai = pyfftw.interfaces.numpy_fft.irfft(Ah, threads=1)
    tend = time.time()

    # error after nmax fft cycles
    abs_err = np.abs((A-Ai)).max()
    rel_err = np.abs((A-Ai)/A).max()

    print("RFFT test with Nx = %i" %Nx)
    print("Performed %i Forward/Inverse FFTs in %f5 seconds, CPU time" %(n,tend-tstart))
    print("Absolute error after %i Forward/Inverse FFTs = %e" %(n,abs_err))
    print("Relative error after %i Forward/Inverse FFTs = %e " %(n,rel_err))
    print(" ")

if __name__ == "__main__":

#    N = 2**np.arange(6,11)
#    for n in N:
#        test_fftw_rfft(Nx = n)

    N = 2**np.arange(5,8)
    for n in N:
        test_fftw_rfft2(Nx = n)



from __future__ import print_function
from future.utils import iteritems
import unittest
import numpy as np
import pyqg
from pyqg import diagnostic_tools as diag

class DiagnosticsTester(unittest.TestCase):

    def test_describe_diagnostics(self):
        """ Test whether describe_diagnostics runs without error """

        m = pyqg.QGModel(1)
        m.describe_diagnostics()

if __name__ == "__main__":
    unittest.main()



from __future__ import print_function
import unittest
import numpy as np
import pyqg

class PyqgModelTester(unittest.TestCase):

    def setUp(self):
        # need to eliminate beta and U for tests
        self.m = pyqg.QGModel(beta=0., U1=0., U2=0., filterfac=0.)
        # the maximum wavelengths to use in tests
        # if we go to higher wavelengths, we don't get machine precision
        self.kwavemax = int(self.m.nx/8)
        self.lwavemax = int(self.m.ny/8)

    def test_fft2(self, rtol=1e-15):
        """Check whether pyqg fft functions produce the expected results."""
        # define a field with a known Fourier transform
        # if I go higher than a factor of 8, the tests start failing
        for kwave in range(1, self.kwavemax):
            for lwave in range(1, self.lwavemax):

                k = 2*np.pi*kwave/self.m.L
                l = 2*np.pi*lwave/self.m.W

                # check whether these are the correct wavenumbers
                np.testing.assert_allclose(self.m.kk[kwave], k, rtol,
                    err_msg='Incorrect wavenumber (kwave=%g)' % kwave)
                np.testing.assert_allclose(self.m.ll[lwave], l, rtol,
                    err_msg='Incorrect wavenumber (lwave=%g)' % lwave)
                np.testing.assert_allclose(self.m.ll[-lwave], -l, rtol,
                    err_msg='Incorrect wavenumber (lwave=%g)' % lwave)


                q1 = np.cos(k * self.m.x )
                q2 = np.sin(l * self.m.y)

                # assign it to the PV - FFT should also get taken at this point
                self.m.set_q1q2(q1, q2)

                # amplitude of RFFT
                qhamp = np.real(self.m.qh * self.m.qh.conj())

                # expected amplitude of RFFT
                amp = (self.m.nx/2)**2 * self.m.ny**2

                # only have one Fourier component for k-axis
                np.testing.assert_allclose(qhamp[0,0,kwave], amp, rtol,
                    err_msg='Incorrect wave amplitude from FFT (kwave=%g)' % kwave)
                # two symmetric pairs for l-axis
                np.testing.assert_allclose(qhamp[1,lwave,0], amp, rtol,
                    err_msg='Incorrect wave amplitude from FFT (lwave=%g)' % lwave)
                np.testing.assert_allclose(qhamp[1,-lwave,0], amp, rtol,
                    err_msg='Incorrect wave amplitude from FFT (lwave=%g)' % lwave)

                # now mask those components
                qhamp_mask = np.ma.masked_array(qhamp, np.zeros_like(qhamp))
                qhamp_mask.mask[0,0,kwave] = 1
                qhamp_mask.mask[1,lwave,0] = 1
                qhamp_mask.mask[1,-lwave,0] = 1
                # and make sure everything else is zero
                np.testing.assert_allclose(qhamp_mask.filled(0.), 0.,
                    rtol=0., atol=rtol,
                    err_msg='Incorrect wave amplitude from FFT')


    def test_inversion_barotropic(self, rtol=1e-13):
        """Check whether inverting a barotropic PV gives desired result.
        Can't get it to work with rtol < 1e-13 """
        # for barotropic, $q = \nabla^2 \psi$
        #                 $\hat{q} = -(k^2 + l^2) \hat \psi$
        #
        # velocity: u = -dpsi/dy, v = dpsi/dx

        for kwave in range(1, self.kwavemax):
            for lwave in range(1, self.lwavemax):

                k = 2*np.pi*kwave/self.m.L
                l = 2*np.pi*lwave/self.m.W

                q =  np.cos(k * self.m.x ) + np.sin(l * self.m.y)
                psi = -k**-2 * np.cos(k * self.m.x ) - l**-2 * np.sin(l * self.m.y)
                u = l**-1 * np.cos(l * self.m.y)
                v = k**-1 * np.sin(k * self.m.x)

                self.m.set_q1q2(q, q)
                self.m._invert()

                for nz in range(self.m.nz):
                    np.testing.assert_allclose(self.m.u[nz], u, rtol,
                        err_msg='Incorrect velocity from barotropic pv inversion')

                    np.testing.assert_allclose(self.m.v[nz], v, rtol,
                        err_msg='Incorrect velocity from barotropic pv inversion')

    def test_inversion_baroclinic(self, rtol=1e-13):
        """Check whether inverting a baroclinic PV gives desired result."""
        # need to think about how to implement this
        pass

    def test_change_inversion_matrix(self):
        """Make sure we can change the inversion matrix after kernel has been
        initialized."""
        a_new = np.random.rand(self.m.nz, self.m.nz, self.m.nl, self.m.nk)
        self.m.a = a_new
        np.testing.assert_allclose(a_new, self.m.a)

    def test_advection(self, rtol=1e-14):
        """Check whether calculating advection tendency gives the descired result."""
        # sin(2 a) = 2 sin(a) cos(a)

        for kwave in range(1, self.kwavemax):
            for lwave in range(1, self.lwavemax):
                k = 2*np.pi*kwave/self.m.L
                l = 2*np.pi*lwave/self.m.W

                q1 =  np.cos(k * self.m.x )
                q2 =  np.sin(l * self.m.y )
                self.m.set_q1q2(q1, q2)
                # manually set velocity
                u1 = np.sin(k * self.m.x)
                v1 = np.zeros_like(self.m.y)
                u2 = np.zeros_like(self.m.x)
                v2 = np.cos(l * self.m.y)
                self.m.u[0] = u1
                self.m.v[0] = v1
                self.m.u[1] = u2
                self.m.v[1] = v2
                self.m.set_U1U2(0.,0.)

                # calculate tendency
                #self.m._advection_tendency()
                self.m._do_advection()
                dqhdt_adv = self.m.dqhdt

                # expected amplitude of RFFT
                amp = (self.m.nx/2)**2 * self.m.ny**2
                tabs = np.real(dqhdt_adv * dqhdt_adv.conj())

                # these tests pass, but what about the factor of two?
                np.testing.assert_allclose(tabs[0,0,2*kwave], k**2 * amp, rtol,
                    err_msg='Incorrect advection tendency k (%g,%g)' % (lwave,kwave))
                np.testing.assert_allclose(tabs[1,2*lwave,0], l**2 * amp, rtol,
                    err_msg='Incorrect advection tendency +l (%g,%g)' % (lwave,kwave))
                np.testing.assert_allclose(tabs[1,-2*lwave,0], l**2 * amp, rtol,
                    err_msg='Incorrect advection tendency -l (%g,%g)' % (lwave,kwave))

                # now mask those components
                tabs_mask = np.ma.masked_array(tabs, np.zeros_like(tabs))
                tabs_mask.mask[0,0,2*kwave] = 1
                tabs_mask.mask[1,2*lwave,0] = 1
                tabs_mask.mask[1,-2*lwave,0] = 1
                # and make sure everything else is zero
                if np.any(np.isnan(tabs_mask.filled(0.))):
                    print("Found NaNs")
                np.testing.assert_allclose(tabs_mask.filled(0.), 0.,
                    rtol=0., atol=rtol,
                    err_msg='Incorrect advection tendency (%g,%g)' % (lwave,kwave))

    def test_friction(self, rtol=1e-15):
        """Check whether calculating advection tendency gives the expected result."""
        # sin(2 a) = 2 sin(a) cos(a)

        for kwave in range(1, self.kwavemax):
            for lwave in range(1, self.lwavemax):
                k = 2*np.pi*kwave/self.m.L
                l = 2*np.pi*lwave/self.m.W

                q1 =  np.cos(k * self.m.x )
                q2 =  np.sin(l * self.m.y )
                self.m.set_q1q2(q1, q2)
                self.m._invert()
                # make sure tendency is zero
                self.m.dqhdt[:] = 0.
                self.m._do_friction()

                # from code
                #self.dqhdt[-1] += self.rek * self.wv2 * self.ph[-1]
                expected = self.m.rek * self.m.wv2 * self.m.ph[-1]
                np.testing.assert_allclose(self.m.dqhdt[-1], expected, rtol,
                        err_msg='Ekman friction was wrong.')
                np.testing.assert_allclose(self.m.dqhdt[:-1], 0., rtol,
                        err_msg='Nonzero friction found in upper layers.')


    def test_timestepping(self, rtol=1e-15):
        """Make sure timstepping works properly."""

        # set initial conditions to zero
        self.m.set_q(np.zeros_like(self.m.q))

        # create a random tendency
        dqhdt = np.random.rand(*self.m.dqhdt.shape) + 1j*np.random.rand(*self.m.dqhdt.shape)
        self.m.dqhdt[:] = dqhdt
        # hack filter to be constant
        #self.m.filtr = 1.

        # make sure we are at the zero timestep
        self.assertEqual(self.m.tc, 0)
        # step forward first time (should use forward Euler)
        self.m._forward_timestep()

        np.testing.assert_allclose(self.m.qh, 1*self.m.dt*dqhdt,
            err_msg='First timestep incorrect')
        # step forward second time (should use AB2)
        self.m._forward_timestep()
        np.testing.assert_allclose(self.m.qh, 2*self.m.dt*dqhdt,
            err_msg='Second timestep incorrect')
        # step forward third time (should use AB3)
        self.m._forward_timestep()
        np.testing.assert_allclose(self.m.qh, 3*self.m.dt*dqhdt,
            err_msg='Third timestep incorrect')



if __name__ == '__main__':
    unittest.main()



from __future__ import print_function
import numpy as np
import pyqg
import xarray as xr
import pytest

year = 24*60*60*360.

expected_vars = [
    'q', 
    'u', 
    'v', 
    'ufull', 
    'vfull', 
    'qh', 
    'uh', 
    'vh', 
    'ph', 
    'Ubg', 
    'Qy',
]

expected_diags = [
    'EKE', 
    'entspec', 
    'APEflux', 
    'APEflux_div', 
    'KEflux', 
    'KEflux_div', 
    'APEgen',
    'APEgenspec',
    'Ensspec',
    'KEspec',
    'EKEdiss',
    'KEspec_modal',
    'PEspec_modal',
    'APEspec',
    'ENSflux',
    'ENSgenspec',
]

expected_attrs = [
    'L',
    'W',
    'dt',
    'filterfac',
    'nk',
    'nl',
    'ntd',
    'nx',
    'ny',
    'nz',
    'rek',
    'taveint',
    'tavestart',
    'tc',
    'tmax',
    'twrite',
]

expected_coords = [
    'time',
    'lev',
    'x',
    'y',
    'l',
    'k',
]

def QG():
    '''Initialize Two-layer Model'''
    return pyqg.QGModel(tmax=year/2, twrite=10000, tavestart=year/3)

def Layered():
    '''Initialize Layered quasigeostrophic model'''
    L =  1000.e3     # length scale of box    [m]
    Ld = 15.e3       # deformation scale      [m]
    kd = 1./Ld       # deformation wavenumber [m^-1]
    Nx = 64          # number of grid points

    H1 = 500.        # layer 1 thickness  [m]
    H2 = 1750.       # layer 2
    H3 = 1750.       # layer 3

    U1 = 0.05          # layer 1 zonal velocity [m/s]
    U2 = 0.025         # layer 2
    U3 = 0.00          # layer 3

    rho1 = 1025.
    rho2 = 1025.275
    rho3 = 1025.640

    rek = 1.e-7       # linear bottom drag coeff.  [s^-1]
    f0  = 0.0001236812857687059 # coriolis param [s^-1]
    beta = 1.2130692965249345e-11 # planetary vorticity gradient [m^-1 s^-1]

    Ti = Ld/(abs(U1))  # estimate of most unstable e-folding time scale [s]
    dt = Ti/200.       # time-step [s]
    tmax = 300*Ti      # simulation time [s]

    return pyqg.LayeredModel(nx=Nx, nz=3, U = [U1,U2,U3],V = [0.,0.,0.],L=L,f=f0,beta=beta,
                             H = [H1,H2,H3], rho=[rho1,rho2,rho3],rek=rek,
                            dt=dt,tmax=tmax, twrite=5000, tavestart=Ti*10)

def BT():
    '''Initialize Barotropic model'''
    return pyqg.BTModel(L=2.*np.pi, nx=256, beta=0., H=1., rek=0., 
                     rd=None, tmax=40, dt=0.001, taveint=1, ntd=4)
    
def SQG():
    '''Initialize Surface Quasi-Geostrophic Model'''
    return pyqg.SQGModel(L=2.*np.pi, nx=512, tmax = 26.005, beta = 0., 
                           Nb = 1., H = 1., rek = 0., rd = None, dt = 0.005,
                           taveint=1, twrite=400, ntd=4)
    
@pytest.fixture(params=[QG, Layered, BT, SQG])
def all_models(request):
    model = request.param()
    return model

def test_xarray(all_models):
    '''Run with snapshots and test contents of xarray.dataset'''
    m=all_models
    tsnapint=year/4
    for snapshot in m.run_with_snapshots(tsnapstart=m.t, tsnapint=tsnapint):
        ds = m.to_dataset()
        assert type(ds) == xr.Dataset

        if snapshot < tsnapint:

            for v in expected_vars:
                assert v in ds

            for a in expected_attrs:
                assert f"pyqg:{a}" in ds.attrs

            for c in expected_coords:
                assert c in ds.coords

        if snapshot > tsnapint:

            for v in list(ds.keys()): 
                assert v in expected_vars + expected_diags

            for a in expected_attrs:
                assert f"pyqg:{a}" in ds.attrs

            for c in expected_coords:
                assert c in ds.coords
